#  ifndef _KDTREE_
#  define _KDTREE_

#  include <iostream>
#  include <fstream>
#  include <string>
#  include <cmath>
#  include <cstdlib>

#  include <bbox.h>

   Real dist( Int, Real *, Int, Real *x[] );

/** A box in a KD tree.
 **/
   class cKdBox:public cBbox
  {
      protected:
         Int parent;                                        /**< Parent box**/
         Int child[2];                                      /**< Children **/
         Int ips,ipe;                                       /**< Range of indexes corresponding to points in the box**/
      public:
         cKdBox();
         cKdBox(Int Parent, Int Child0, Int Child1, Int Ips, Int Ipe ); 
         cKdBox(Real X0[], Real X1[], Int Parent, Int Child0, Int Child1, Int Ips, Int Ipe ); 
         void range( Int *Ips, Int *Ipe ){ *Ips=ips; *Ipe=ipe; };
         void corners( Real X0[], Real X1[] );
         void setChildren( Int Child0, Int Child1 ){ child[0]=Child0;child[1]= Child1; };
         void getChildren( Int *Child0, Int *Child1 ){ *Child0=child[0]; *Child1=child[1]; };
         Int getParent( ){ return parent; };
         void plot();
         Real dist( Int, Real * );
         Real dist( Int, Int, Real *[] );
  };


   class cKdTree
  {
      protected:
         Int            nx;                                  /**< number of space directions**/
         Int            np;                                  /**< Number of points **/
         Int            nbxs;                                /**< Number of boxes **/
         cKdBox        *bxs;
         Int           *iprm,*iprmi;
         Real          *xp[3];

//       void sift_down( Int , Real *, Int * );

      public:
         cKdTree();
         void build( Int Nx,Int Np, Real *x[] );

         Int locate( Real * );
         Int locate( Int );
         Int  nearest( Real * );

         #pragma acc routine seq
         void nearest( Real *, Int *, Real * );

         void print( Int );
         Real dist( Int, Int );
         Real dist( Real *, Int );
         void clean();
        ~cKdTree();
  };

   template <typename type > void swap( type *v0, type *v1 ){ type dum; dum= *v0; *v0= *v1; *v1= dum; };


   template < typename type > void sift_down( Int n, type *var, Int *iprm )
  {
      Int m= n-1;

      Int j,j0;
      Int  ia;
      type a;

      a= var[0];
      ia= iprm[0];
      j0= 0;
      j= 1;
      while( j<= m )
     {
         if( j < m && var[j] < var[j+1] ) j++;
         if( a >= var[j] ) break;
         var[j0]= var[j];
         iprm[j0]= iprm[j];
         j0= j;
         j= 2*j+1;
     }
      var[j0]= a;
      iprm[j0]= ia;
  }

   template < typename type > void select( Int k, Int n, Int *iprm, type *x )
  {
      Int i,ia,ir,j,l,im;
      Int iswp;
      type   a;

      iswp=0;
      l= 0;
      ir=n-1;
      for(;;)
     {
         iswp++;
         if( ir<= l+1 )
        {
            if( ( ir == l+1 ) && ( x[iprm[ir]] < x[iprm[l]] ) )
           {
               swap( iprm+l,iprm+ir );
           }
            return;
        }
         else
        {
            im= ( l+ir ) >> 1;
            swap( iprm[im],iprm[l+1] );
            if( x[iprm[l]] > x[iprm[ir]] ) swap( iprm+l,iprm+ir );
            if( x[iprm[l+1]] > x[iprm[ir]] ) swap( iprm+l+1,iprm+ir );
            if( x[iprm[l]] > x[iprm[l+1]] ) swap( iprm+l,iprm+l+1 );
            i= l+1;
            j= ir;
            ia= iprm[l+1];
            a= x[ia];
            for(;;)
           {
               do i++; while( x[iprm[i]] < a );
               do j--; while( x[iprm[j]] > a );
               if( j < i )break;
               swap( iprm+i,iprm+j );
           }
            iprm[l+1]= iprm[j];
            iprm[j]= ia;
            if( j>= k ) ir=j-1;
            if( j<= k ) l=i;
        }
     }
  }

#  endif
